import fs from "fs";
import jwt from "jsonwebtoken";
import crypto from "crypto";
import axios, { AxiosInstance } from "axios";
import NodeCache from "node-cache";
import { match } from "path-to-regexp";
import { RequestHandler } from "express";

export enum Methods { POST = 'POST', PUT = 'PUT', DELETE = 'DELETE', PATCH = 'PATCH', GET = 'GET' }

export type RoutePermission = { method: Methods; path: string };
export type Scopes = { [domain: string]: RoutePermission[] };
export type JwtPayload = { service_from: string; service_to: string; scope: RoutePermission[] };
export type ServiceMeta = { clientPrivateKey: string; portalPublicKey: string; serviceId: string };

export type PortalOptions = {
    hostname?: string;
    portalHost?: string;
    metaFilePath?: string;
    clientId: string;
    clientSecret: string;
    label?: string;
    description?: string;
};

export class CrysteamPortal {
    private _serviceMeta?: ServiceMeta;
    private readonly _tokenCache = new NodeCache({ stdTTL: 1800 });

    private readonly _axios: AxiosInstance;
    private readonly _hostname?: string;
    private readonly _portalHost: string;
    private readonly _metaFilePath: string;
    private readonly _secretKey: string;

    constructor(options: PortalOptions) {
        this._portalHost = options.portalHost || "https://portal.crysteam.ru/api/v2";
        this._metaFilePath = options.metaFilePath || "./data/.portal";
        this._hostname = options.hostname;

        this._secretKey = Buffer.from(`${options.clientId}:${options.clientSecret}`).toString("base64");
        this._axios = axios.create({
            baseURL: this._portalHost,
            headers: { authorization: `Bearer ${this._secretKey}` }
        });

        this._initialize(options)
    }

    private async _initialize(options: PortalOptions) {
        if(options.label)
            await this.registerService(options.label, options.description);
    }

    private sign(text: string): string {
        if (!this._serviceMeta) throw new Error("[Portal] Meta not loaded");
        const signer = crypto.createSign("RSA-SHA256");
        signer.update(text);
        signer.end();
        return signer.sign(this._serviceMeta.clientPrivateKey, "base64");
    }

    private buildSignedPayload(data: object = {}): object {
        if (!this._serviceMeta) throw new Error("[Portal] Meta not loaded");
        const signature = this.sign(JSON.stringify(data));
        return {
            service_id: this._serviceMeta.serviceId,
            signature,
            payload: data,
        };
    }

    private generateRsaKeyPair() {
        console.log("[Portal] Generating RSA key pair...");
        return crypto.generateKeyPairSync("rsa", {
            modulusLength: 2048,
            publicKeyEncoding: { type: "spki", format: "pem" },
            privateKeyEncoding: { type: "pkcs8", format: "pem" },
        });
    }

    private saveMetaToFile(meta: ServiceMeta) {
        console.log(`[Portal] Saving meta to ${this._metaFilePath}`);
        fs.mkdirSync(require("path").dirname(this._metaFilePath), { recursive: true });
        fs.writeFileSync(this._metaFilePath, JSON.stringify(meta, null, 2), "utf8");
    }

    private loadMetaFromFile(): ServiceMeta | undefined {
        console.log(`[Portal] Loading meta from ${this._metaFilePath}`);
        if (!fs.existsSync(this._metaFilePath)) return;
        try {
            return this._serviceMeta = JSON.parse(fs.readFileSync(this._metaFilePath, "utf8"));
        } catch (err) {
            console.error("[Portal] Failed to parse meta:", err);
        }
    }

    async getScopes(): Promise<Scopes> {
        const { data } = await this._axios.post("/services/scopes", this.buildSignedPayload());
        return data;
    }

    async registerService(label: string, description?: string): Promise<string | undefined> {
        if (this.loadMetaFromFile() || this._serviceMeta) return;

        const { privateKey, publicKey } = this.generateRsaKeyPair();
        const { data } = await this._axios.post("/services/register", {
            label,
            description,
            publicKey,
            hostname: this._hostname,
        });

        const meta: ServiceMeta = {
            clientPrivateKey: privateKey,
            portalPublicKey: data.publicKey,
            serviceId: data.id,
        };

        this.saveMetaToFile(meta);
        this._serviceMeta = meta;
        return data.id;
    }

    async generateAccessToken(serviceFrom: string, scope: RoutePermission[]): Promise<string> {
        if (!this._serviceMeta) throw new Error("[Portal] Meta not loaded");

        const cacheKey = `${serviceFrom}:${JSON.stringify(scope)}`;
        const cached = this._tokenCache.get<{ token: string }>(cacheKey);
        if (cached) return cached.token;

        const tokenPayload = {
            service_from: serviceFrom,
            service_to: this._serviceMeta.serviceId,
            scope,
        };

        const signature = this.sign(JSON.stringify(tokenPayload));
        console.log("[Portal] Requesting access token...");

        const { data: token } = await this._axios.post("/services/produce", {
            ...tokenPayload,
            signature,
        });

        this._tokenCache.set(cacheKey, { token });
        return token;
    }

    async getServiceScopesFor(domain: string): Promise<RoutePermission[]> {
        const scopes = await this.getScopes();
        const domainScopes = scopes[domain];
        if (!domainScopes) throw new Error(`[Portal] No scopes found for domain "${domain}"`);
        return domainScopes;
    }

    isScopeAllowed(scopes: RoutePermission[], permission: RoutePermission): boolean;
    isScopeAllowed(domain: string, permission: RoutePermission): Promise<boolean>;
    isScopeAllowed(arg: string | RoutePermission[], permission: RoutePermission): boolean | Promise<boolean> {
        if (Array.isArray(arg)) {
            return arg.some(scope =>
                scope.method === permission.method &&
                scope.path === permission.path
            );
        }

        return this.getServiceScopesFor(arg)
            .then(scopes =>
                scopes.some(scope =>
                    scope.method === permission.method &&
                    scope.path === permission.path
                )
            )
            .catch(() => false);
    }

    createHttpClientFor(serviceFrom: string, scope: RoutePermission[]) {
        const client = axios.create({ baseURL: serviceFrom });

        client.interceptors.request.use(async config => {
            const token = await this.generateAccessToken(serviceFrom, scope);
            config.headers['x-service-token'] = `Bearer ${token}`;
            return config;
        });

        client.interceptors.response.use(undefined, async error => {
            if (error.response?.status === 403) {
                const token = await this.generateAccessToken(serviceFrom, scope);
                error.config.headers['x-service-token'] = `Bearer ${token}`;
                return axios.request(error.config);
            }
            return Promise.reject(error);
        });

        return client;
    }

    verifyApiKey(): RequestHandler {
        return (req, res, next) => {
            try {
                if (!this._hostname) throw new Error("Hostname is not configured.");
                const fullPatch = req.baseUrl + req.path;
                console.log(`[Portal][Auth] ${req.method} ${fullPatch}`);

                const serviceToken = req.headers['x-service-token'] as string | undefined;
                if (!serviceToken) throw new Error("Missing token.");

                const token = serviceToken.split(" ")[1];
                if (!token) throw new Error("Missing token.");

                const meta = this.loadMetaFromFile();
                if (!meta) throw new Error("Missing meta information.");

                const decoded = jwt.verify(token, meta.portalPublicKey, {
                    algorithms: ["RS256"],
                }) as JwtPayload;

                const isAllowed = decoded.scope.some(({ method, path }) => {
                    if (decoded.service_from !== this._hostname) return false;
                    if (method !== req.method) return false;
                    return match(path, { decode: decodeURIComponent })(fullPatch) !== false;
                });

                if (!isAllowed) throw new Error("Access denied.");

                (req as any).user = decoded;
                next();
            } catch (err) {
                next(err);
            }
        };
    }
}
